"""
This package contains classes for representation of HDL
languages (hdl types, RtlSignal, Assignment, HWProcess etc.).

It also contains classes for represenation of complex HW strucures
like dense transaction template (TransTmpl, FrameTmpl).
"""
