VHLD_KEYWORDS = [
    "abs", "access", "across", "after", "alias", "all", "and", "architecture",
    "array", "assert", "attribute", "begin", "block", "body", "break",
    "bugger", "bus", "case", "component", "configuration", "constant",
    "disconnect", "downto", "end", "entity", "else", "elsif", "exit", "file",
    "for", "function", "generate", "generic", "group", "guarded", "if",
    "impure", "in", "inertial", "inout", "is", "label", "library", "limit",
    "linkage", "literal", "loop", "map", "mod", "nand", "nature", "new",
    "next", "noise", "nor", "not", "null", "of", "on", "open", "or", "others",
    "out", "package", "port", "postponed", "process", "procedure",
    "procedural", "pure", "quantity", "range", "reverse_range", "reject",
    "rem", "record", "reference", "register", "report", "return", "rol", "ror",
    "select", "severity", "shared", "signal", "sla", "sll", "spectrum", "sra",
    "srl", "subnature", "subtype", "terminal", "then", "through", "to",
    "tolerance", "transport", "type", "unaffected", "units", "until", "use",
    "variable", "wait", "with", "when", "while", "xnor", "xor"]
